#lang racket
(require racket/stxparam
         (for-syntax syntax/parse)
         "record-macro.rkt"
         "method-macro.rkt")

(provide class new get set send ssend)

(struct classV (constructor methods))
(struct objectV (class fields))

(define-syntax (class stx)
  (syntax-parse stx
    [{class name:id
       {fld:id ...}
       {mthd-name:id mthd-expr} ...}
     #'(define name
         (classV
          (lambda (fld ...)
            (make-hash (list (cons 'fld fld)
                             ...)))
          (record [mthd-name (method mthd-expr)]
                  ...)))]))

(define-syntax (new stx)
  (syntax-parse stx
    [{new cls:id expr ...}
     (syntax/loc stx
       (objectV
        cls
        ((classV-constructor cls) expr ...)))]))

(define-syntax (get stx)
  (syntax-parse stx
    [{get obj-expr fld:id}
     (syntax/loc stx 
      (hash-ref (objectV-fields obj-expr) 'fld))]))

(define-syntax (set stx)
  (syntax-parse stx
    [{set obj-expr fld:id arg-expr}
     (syntax/loc stx
       (let ([obj obj-expr])
         (hash-set! (objectV-fields obj) 'fld arg-expr)
         obj))]))

(define-syntax (send stx)
  (syntax-parse stx
    [{send obj-expr mthd:id arg-expr}
     (syntax/loc stx
       (let ([obj obj-expr])
         ((hash-ref (classV-methods (objectV-class obj)) 'mthd)
          obj
          arg-expr)))]))

(define-syntax (ssend stx)
  (syntax-parse stx
    [{ssend obj-expr cls:id mthd:id arg-expr}
     (syntax/loc stx
       (let ([obj obj-expr])
         ((hash-ref (classV-methods cls) 'mthd)
          obj
          arg-expr)))]))

(module+ test
  (require rackunit)
  
  {class posn
    {x y}
    {mdist {+ {get this y}
              {get this x}}}}
  {class posn3D
    {x y z}
    {mdist {+ {get this z}
              {ssend this posn mdist arg}}}}

  (check-equal? {send {new posn 1 2}
                      mdist
                      0}
                3)
  (check-equal? {send {new posn3D 1 2 3}
                      mdist
                      0}
                6))
