#lang racket
(require (for-syntax syntax/parse racket/list))

(provide record)

#|
(record [x 10]
        [y 20])

(get ... x)
(set ... x ...)
|#

(define-syntax record
  (lambda (a) 
    (syntax-parse a
     [(record [x n] ...)
      (define  xs (syntax-e #'(x ...)))
      (map (lambda (x) (if (symbol? (syntax-e x))
                           #t
                           (raise-syntax-error 
                            'record
                            "not a symbol"
                            x)))
           xs)
      ;; Same thing, more Rackety:
      #;
      (for ([x xs])
        (unless (symbol? (syntax-e x))
          (raise-syntax-error 
           'record
           "not a symbol"
           x)))
      (define l (interleave (syntax-e #'('x ...)) (syntax-e #'(n ...))))
      ;; Same as below: (datum->syntax #'here (cons #'hash l))
      #`(hash #,@l)])))

(begin-for-syntax
  (define (interleave l r)
    (cond
      [(empty? l) empty]
      [(cons? l) (cons (first l) (cons (first r) (interleave (rest l) (rest r))))])) )

(module+ test
  (require rackunit)
  (check-equal? (record [x 2]
                        [y 3])
                (hash 'x 2 'y 3)))
  
